from math import radians, cos, sin, asin, sqrt

def calculate_distance_based_on_lon_lat(dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    https://en.wikipedia.org/wiki/Great-circle_distance
    The current algorothm is the  haversine formula 

    Args:
        dLongitude_from (float): The longitude of the start point
        dLatitude_from (float):  The latitude of the start point
        dLongitude_to (float):  The longitude of the end point
        dLatitude_to (float):  The latitude of the end point

    Returns:
        float: The great circle distance
    """
    # convert decimal degrees to radians 
    dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to = map(radians, [dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to])

    # haversine formula 
    dlon = dLongitude_to - dLongitude_from 
    dlat = dLatitude_to - dLatitude_from 
    a = sin(dlat/2)**2 + cos(dLatitude_from) * cos(dLatitude_to) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a)) 
    r = 6378137.0# Radius of earth in kilometers. Use 3956 for miles
    return c * r


import os, sys
from osgeo import gdal, osr,ogr, gdalconst

def gdal_write_geotiff_file(sFilename_in, aData_in, dPixelWidth_in,    dOriginX_in,        dOriginY_in,    dMissing_value_in,     pSpatial_reference_in):
    """
    Write a Geotiff standard format raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName='GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 

    nrow, ncolumn = aData_in.shape
    nband = 1

    pDataset = pDriver.Create(
        sFilename_in,
        ncolumn,
        nrow,
        nband,
        gdal.GDT_Float32, )

    pDataset.SetGeoTransform([
        dOriginX_in,    # 0
        dPixelWidth_in,  # 1
        0,                      # 2
        dOriginY_in,    # 3
        0,                      # 4
        -dPixelWidth_in])

    pProjection= pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    pBand = pDataset.GetRasterBand(1)
    pBand.WriteArray(aData_in)
    pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver = None
    pDataset = None
    pBand=None
    return  sFilename_in, pGeotransform_out, pProjection_out




def gdal_write_geotiff_file_multiple_band(sFilename_in, aData_in,  dPixelWidth_in,   dOriginX_in,   dOriginY_in, dMissing_value_in,   pSpatial_reference_in ):
    """
    Write a multi-band geotiff raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName='GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 

    nband, nrow, ncolumn = aData_in.shape


    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in, \
                              ncolumn, \
                              nrow, \
                              nband,\
                              gdal.GDT_Float32)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in, \
                              dPixelWidth_in, \
                              0.0, \
                              dOriginY_in, \
                              0.0, \
                              -dPixelWidth_in])

    pProjection= pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    #Write raster datasets
    for iBand in range(nband):
        pBand = pDataset.GetRasterBand(iBand + 1)
        pBand.WriteArray(aData_in[iBand, :, :])
        pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver =None
    pDataset=None
    pBand=None

    return  sFilename_in, pGeotransform_out, pProjection_out




import math as Math
import numpy as np

def calculate_ticks_space (aData, nstep_in = 5, iFlag_small_in = None):
    """
    Computes domain with given nstep_in encompassing series aData
    @ params
    aData    - Required - A list-like object of integers or floats
    nstep_in - Optional - Tick frequency
    """

    if iFlag_small_in is not None:
        iFlag_small = 1
    else:
        iFlag_small = 0
    if iFlag_small == 1:
        xMax = np.nanmax(aData) 
        xMin =  np.nanmin(aData) 
        dSpace = (xMax - xMin)/ (nstep_in-2)
        dMax = xMax * 1.2
        dMin =  xMin * 0.8

    else: 
        
        xMax = Math.ceil( np.nanmax(aData) )
        xMin = Math.floor( np.nanmin(aData) )
      
        dSpace = (xMax - xMin)/ (nstep_in-2)

        dMax = xMax + Math.ceil( dSpace )
        dMin =  Math.floor( xMin - dSpace) 
    return (dSpace, dMin, dMax)


import os, sys
import numpy as np
import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.ticker as ticker
from datetime import datetime

from pyearth.system.define_global_variables import *



def plot_time_series_data_with_two_y_axis(aTime_all, aData_all, \
                                  sFilename_out,\
                                  iDPI_in = None,\
                                  iFlag_trend_in = None, \
                                  iReverse_y_in = None, \
                                  iSize_x_in = None, \
                                  iSize_y_in = None, \
                                  aMax_y_in =None, \
                                  aMin_y_in = None, \
                                  aSpace_y_in=None,\
                                  aMarker_in =None,\
                                aColor_in =None,\
                                aLinestyle_in =None,\
                                  aLabel_y_in = None, \
                                  aLabel_legend_in = None,\
                                  sTitle_in = None):
    nData = len(aData_all)
    aData = [aData_all]
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_trend_in is not None:
        iFlag_trend = 1
    else:
        iFlag_trend = 0

    if iReverse_y_in is not None:
        iReverse_y = 1
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12
    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if aLabel_y_in is not None:
        sLabel_y1 = aLabel_y_in[0]
        sLabel_y2 = aLabel_y_in[1]
    else:
        sLabel_y1 = ''
        sLabel_y2 = ''
    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
        sLabel_legend1 = aLabel_legend_in[0]
        sLabel_legend2 = aLabel_legend_in[1]
    else:
        sLabel_legend1 = ''
        sLabel_legend2 = ''
    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''
    
    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker=np.full(nData, '+')

    if aColor_in is not None:
        aColor = aColor_in
    else:
        aColor= np.full(nData, 'black')
       
    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle = np.full(nData, '-')
     

    if aMax_y_in is not None:
        dMax_y1 = aMax_y_in[0]
        dMax_y2 = aMax_y_in[1]
    else:
        dMax_y1  = np.nanmax(aData[0]) * 1.2
        dMax_y2  = np.nanmax(aData[1]) * 1.2
    if aMin_y_in is not None:
        dMin_y1 = aMin_y_in[0]
        dMin_y2 = aMin_y_in[1]
    else:
        #dMin_Y = np.nanmin(aData) #if it has negative value, change here
        dMin_y1  = np.nanmin(aData[0])
        dMin_y2  = np.nanmin(aData[1]) 

    #if (dMax_Y <= dMin_Y ):
    #    return

    if aSpace_y_in is not None:        
        dSpace_y1 = aSpace_y_in[0]
        dSpace_y2 = aSpace_y_in[1]
    else:       
        pass

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x)
    fig.set_figheight( iSize_y)
    ax1 = fig.add_axes([0.1, 0.5, 0.8, 0.4] )
    pYear = mdates.YearLocator(1)   # every year
    pMonth = mdates.MonthLocator()  # every month
    sYear_format = mdates.DateFormatter('%Y')
    #plot the first axis
    i=1
    x1 = aTime_all[i-1]
    
    y1 = aData_all[i-1]
    ax1.plot( x1, y1, \
            color = aColor[i-1], linestyle = aLinestyle[i-1] ,\
            marker = aMarker[i-1] , markersize =0.5,\
            label = aLabel_legend[i-1])
       
    #calculate linear regression
    nan_index = np.where(y1 == missing_value)
    y1[nan_index] = np.nan
    good_index = np.where(  ~np.isnan(y1))
    if iFlag_trend ==1:
        x_dummy = np.array( [i.timestamp() for i in x1 ] )
        x_dummy = x_dummy[good_index]
        y_dummy = y1[good_index]
        coef = np.polyfit(x_dummy,y_dummy,1)
        poly1d_fn = np.poly1d(coef)
        mn = np.min(x_dummy)
        mx = np.max(x_dummy)
        x2 = [mn,mx]
        y2 = poly1d_fn(x2)
        x2 = [datetime.fromtimestamp(i) for i in x2 ]
        ax1.plot(x2,y2, color = 'orange', linestyle = '-.',  linewidth=0.5)
    #plot the second axis
    ax2 = ax1.twinx()
    i=2
    x2 = aTime_all[i-1]
    y2 = aData_all[i-1]
    ax2.plot( x2, y2, \
            color = aColor[i-1], linestyle = aLinestyle[i-1] ,\
            marker = aMarker[i-1] ,\
            label = aLabel_legend[i-1])


    ax1.axis('on')
    ax1.grid(which='major', color='grey', linestyle='--', axis='y')
    
    #ax1.set_aspect(dRatio)  #this one set the y / x ratio
    ax1.xaxis.set_major_locator(pYear)
    ax1.xaxis.set_minor_locator(pMonth)
    ax1.xaxis.set_major_formatter(sYear_format)
    ax1.tick_params(axis="x", labelsize=10)
    ax1.tick_params(axis="y", labelsize=10)

    ax1.set_xmargin(0.05)
    ax1.set_ymargin(0.15)

    ax1.set_xlabel('Year',fontsize=12)
    ax1.set_ylabel(sLabel_y1,fontsize=12)
    ax2.set_ylabel(sLabel_y2,fontsize=12)
    ax1.set_title( sTitle, loc='center', fontsize=15)
    # round to nearest years...
    x_min = np.datetime64(aTime_all[1][0], 'Y') 
    x_max = np.datetime64(aTime_all[1][ -1 ], 'Y') + np.timedelta64(1, 'Y')
    ax1.set_xlim(x_min, x_max)
    if dMax_y1 < 1000 and dMax_y1 > 0.001:
        ax1.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1f'))
    else:
        ax1.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1e'))
        
    
    ax1.yaxis.set_major_locator(ticker.MultipleLocator(dSpace_y1))
    ax2.yaxis.set_major_locator(ticker.MultipleLocator(dSpace_y2))
    if (iReverse_y == 1):
        ax1.set_ylim( dMax_y1, dMin_y1 )
    else:
        ax1.set_ylim( dMin_y1, dMax_y1 )
        
    ax1.legend(bbox_to_anchor=(1.0,1.0), loc="upper right", fontsize=12)
    ax2.legend(bbox_to_anchor=(1.0,1.0), loc="upper right", fontsize=12)
    print(sFilename_out)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    #print('finished plotting')



import os, sys
import numpy as np
from pyevtk.hl import  unstructuredGridToVTK
from pyevtk.vtk import  VtkQuad

#convert a point based data to unstructured polygon vtk 
def convert_array_to_vtk_polygon(pArray_in, pLabel_in, aX_in, aY_in, pZ_in, sFilename_vtk):
    

    dMin_x = np.min(aX_in)
    dMax_x = np.max(aX_in)

    dMin_y = np.min(aY_in)
    dMax_y = np.max(aY_in)
    pShape_x = aX_in.shape
    #get resolution in x
    dResolution_x = ( dMax_x - dMin_x) / (pShape_x[1] -1 )
    pShape_y = aY_in.shape
    #get resolution in y
    dResolution_y = ( dMax_y - dMin_y ) / (pShape_y[0] -1 )
    pShape_z = pZ_in.shape
    #center
    nrow_center = pShape_y[0]
    ncolumn_center = pShape_x[1]
    nlayer_center = pShape_z[0] - 1
    #vertex
    nrow_vertex = nrow_center + 1
    ncolumn_vertex = ncolumn_center + 1
    nlayer_vertex = pShape_z[0] #including the base/reference 
    x = np.zeros(( nlayer_vertex , nrow_vertex, ncolumn_vertex))
    y = np.zeros(( nlayer_vertex , nrow_vertex, ncolumn_vertex))
    z = np.full( ( nlayer_vertex , nrow_vertex, ncolumn_vertex), np.nan, dtype=float )


    X = np.arange( dMin_x - 0.5 * dResolution_x, dMax_x + dResolution_x, dResolution_x, dtype='float64')
    Y = np.arange( dMax_y + 0.5 * dResolution_y, dMin_y - dResolution_y, -dResolution_y, dtype='float64')
    aGrid_x, aGrid_y = np.meshgrid(X, Y)


    iFlag =1
    for iLayer in np.arange(1, nlayer_vertex+1, 1):
        #x and y are ready 
        x[iLayer-1, :,: ] = aGrid_x
        y[iLayer-1, :,: ] = aGrid_y

        aZ = np.full( ( nrow_vertex, ncolumn_vertex), np.nan, dtype=float )

        aZ_subset = pZ_in[iLayer-1]
        #now deal with edges, there are two possible arrangments
        if iFlag ==1 : #this is a easy approach
            aZ[0:pShape_y[0],0:pShape_x[1] ] = aZ_subset
            #right edge
            for iRow in np.arange(0, nrow_center, 1 ):
                aZ[iRow, ncolumn_vertex-1] = aZ_subset[iRow, ncolumn_vertex-2]
                pass
            #bottm edge
            for iColumn in np.arange(0, ncolumn_center, 1 ):
                aZ[nrow_vertex-1, iColumn] = aZ_subset[nrow_vertex-2, iColumn]
            #lower right 
            aZ[nrow_center ,ncolumn_center  ] = aZ_subset[nrow_vertex-2, ncolumn_vertex-2]           
            pass
        else:
            #in this case, we set the origin at lower left
            pass

        #set z
        z[iLayer-1,:,:] = aZ     

        pass
       
    # Define connectivity or vertices that belongs to each element
    conn = np.full(( nrow_center * ncolumn_center * nlayer_center * 4), 0, dtype=int)

    #without bottom layer
    for iLayer in np.arange(1, nlayer_center + 1, 1):
        start_id = (iLayer-1) * nrow_vertex * ncolumn_vertex
        start_index = (iLayer-1) * nrow_center * ncolumn_center * 4

        for iRow in np.arange(1, nrow_center+1, 1):
            for iColumn in np.arange(1, ncolumn_center+1, 1 ):
                
                new_index = (iRow-1) * ncolumn_center * 4 + (iColumn-1) * 4

                conn[start_index + new_index] = (iRow-1) * ncolumn_vertex + iColumn-1  + start_id
                conn[start_index + new_index+1] = (iRow-1) * ncolumn_vertex + iColumn + start_id
                conn[start_index + new_index+2] = iRow * ncolumn_vertex + iColumn + start_id
                conn[start_index + new_index+3] = iRow * ncolumn_vertex + iColumn-1 + start_id

                pass  
        pass

    #we don not consider bottom base layer connectivity

    # Define offset of last vertex of each element
    offset = np.arange( 4, (4*ncolumn_center*nrow_center*nlayer_center+1), 4 )    
    # Define cell types
    ctype = np.full( (nrow_center, ncolumn_center, nlayer_center), VtkQuad.tid )


    x.shape =  (nrow_vertex) * (ncolumn_vertex) * (nlayer_vertex)
    y.shape =  (nrow_vertex) * (ncolumn_vertex) * (nlayer_vertex)
    z.shape =  (nrow_vertex) * (ncolumn_vertex) * (nlayer_vertex)

    celldata = {}
    nCelldata = len(pLabel_in)
    ncell = nrow_center * ncolumn_center * nlayer_center

    for i in np.arange(nCelldata):
        sLabel = pLabel_in[i]
        #reshape 
        data=pArray_in[i]
        data.shape = ncell
        celldata[sLabel] = data
        pass
    

    conn= np.reshape(conn, (4*nrow_center*ncolumn_center*nlayer_center))
    #save as vtk file


    unstructuredGridToVTK(sFilename_vtk, x, y, z, connectivity = conn, \
                              offsets = offset, cell_types = ctype, cellData = celldata, pointData = None)



    return



import pdb
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import cartopy.crs as ccrs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER

pProjection = ccrs.PlateCarree()
def map_raster_data(aImage_in, \
    aImage_extent, \
    sFilename_output_in,\
    sColormap_in = None,\
    iDPI_in = None,\
    dMissing_value_in=None,\
    dData_max_in = None, \
    dData_min_in = None):

    aImage_in = np.array(aImage_in)

    pShape = aImage_in.shape
    nrow, ncolumn = aImage_in.shape
    iSize_x = ncolumn
    iSize_y = nrow 
    sFilename_out= sFilename_output_in
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if dMissing_value_in is not None:
        dMissing_value = dMissing_value_in
    else:
        dMissing_value= np.nanmin(aImage_in)

    dummy_index = np.where(aImage_in == dMissing_value)
    aImage_in[dummy_index] = np.nan    

    if dData_max_in is not None:
        dData_max = dData_max_in
    else:
        dData_max = np.nanmax(aImage_in)

    if dData_min_in is not None:
        dData_min = dData_min_in
    else:
        dData_min = np.nanmin(aImage_in)    
    

    if sColormap_in is not None:
        sColormap = sColormap_in
    else:
        sColormap =  'rainbow'

    cmap = cm.get_cmap(sColormap)
   

    dummy_index = np.where(aImage_in > dData_max)
    aImage_in[dummy_index] = dData_max

    dummy_index = np.where(aImage_in < dData_min)
    aImage_in[dummy_index] = dData_min


    fig = plt.figure( dpi = iDPI  )
    #fig.set_figwidth( iSize_x )
    #fig.set_figheight( iSize_y )
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.7], projection=pProjection )


    # set a margin around the data
    ax.set_xmargin(0.05)
    ax.set_ymargin(0.10)   

    ax.imshow(aImage_in, origin='upper', \
        extent=aImage_extent, \
        cmap = cmap, \
        transform=pProjection)   

    ax.coastlines(color='black', linewidth=1)
    #ax.set_global()
    #ax.set_xticks([-180, -120, -60, 0, 60, 120, 180], crs=ccrs.PlateCarree())
    #ax.set_yticks([-90, -60, -30, 0, 30, 60, 90], crs=ccrs.PlateCarree())    
    #lon_formatter = LongitudeFormatter(number_format='.1f',
    #                                   degree_symbol='',
    #                                   dateline_direction_label=True)
    #lat_formatter = LatitudeFormatter(number_format='.1f',
    #                                  degree_symbol='')
    #ax.xaxis.set_major_formatter(lon_formatter)
    #ax.yaxis.set_major_formatter(lat_formatter)
    ax.set_extent(aImage_extent)
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                  linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
    gl.xlabel_style = {'size': 15, 'color': 'gray'}
    gl.xlabel_style = {'color': 'red', 'weight': 'bold'}

    plt.savefig(sFilename_out , bbox_inches='tight')
    #.show()

    plt.close('all')
    plt.clf()

    
    





from math import cos, sin,  sqrt, pi
import numpy as np
def calculate_polygon_area(aLongitude_in, aLatitude_in,  iFlag_algorithm = 0, radius = 6378137.0):
    """
    Computes area of spherical polygon, assuming spherical Earth. 
    Returns result in ratio of the sphere's area if the radius is specified. Otherwise, in the units of provided radius.
    lats and lons are in degrees.

    Args:
        aLongitude_in (list): The longitude of list of points
        aLatitude_in (list): The latitude of list of points
        iFlag_algorithm (int, optional): Which algorithm to use. Defaults to 0.
        radius (float, optional): The radius of Earth in meter. Defaults to 6378137.0.

    Returns:
        float: The area of the polygon
    """
    
    npoint = len(aLongitude_in)
    if npoint<3:
        print('More than 2 points are required!')
        return
   
    #TODO: take into account geodesy (i.e. convert latitude to authalic sphere, use radius of authalic sphere instead of mean radius of spherical earth)
    aLatitude_in = np.deg2rad(aLatitude_in)
    aLongitude_in = np.deg2rad(aLongitude_in)

    if iFlag_algorithm==0:
        # Line integral based on Green's Theorem, assumes spherical Earth
        

        #close polygon
        if aLatitude_in[0]!=aLatitude_in[-1]:
            aLatitude_in = np.append(aLatitude_in, aLatitude_in[0])
            aLongitude_in = np.append(aLongitude_in, aLongitude_in[0])

        # Get colatitude (a measure of surface distance as an angle)
        a = sin(aLatitude_in/2)**2 + cos(aLatitude_in)* sin(aLongitude_in/2)**2
        colat = 2*np.arctan2( sqrt(a), sqrt(1-a) )

        #azimuth of each point in segment from the arbitrary origin
        az = np.arctan2(cos(aLatitude_in) * sin(aLongitude_in), sin(aLatitude_in)) % (2*pi)

        # Calculate step sizes
        # daz = np.diff(az) % (2*pi)
        daz = np.diff(az)
        daz = (daz + pi) % (2 * pi) - pi

        # Determine average surface distance for each step
        deltas=np.diff(colat)/2
        colat=colat[0:-1]+deltas

        # Integral over azimuth is 1-cos(colatitudes)
        integrands = (1-cos(colat)) * daz

        # Integrate and save the answer as a fraction of the unit sphere.
        # Note that the sum of the integrands will include a factor of 4pi.
        area = abs(sum(integrands))/(4*pi) # Could be area of inside or outside

        area = min(area,1-area)
        if radius is not None: #return in units of radius
            return area * 4*pi*radius**2
        else: #return in ratio of sphere total area
            return area
    elif iFlag_algorithm==2:
        #L'Huilier Theorem, assumes spherical earth
        #see:
        # https://mathworld.wolfram.com/SphericalPolygon.html
        # https://web.archive.org/web/20160324191929/http://forum.worldwindcentral.com/showthread.php?20724-A-method-to-compute-the-area-of-a-spherical-polygon
        # https://github.com/spacetelescope/spherical_geometry/blob/master/spherical_geometry/polygon.py
        # https://github.com/tylerjereddy/spherical-SA-docker-demo/blob/master/docker_build/demonstration.py
        #TODO
        pass
    elif iFlag_algorithm==3:
        #https://trs.jpl.nasa.gov/handle/2014/41271
        #TODO
        pass


import os, sys
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.patches as mpl_patches
from scipy.stats import gaussian_kde

import matplotlib.lines as mlines
import matplotlib.transforms as mtransforms

from pyearth.visual.scatter.scatter_lowess import scatter_lowess

from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.visual.color.choose_n_color import polylinear_gradient, rand_hex_color

def scatter_plot_multiple_data(aData_x, \
                      aData_y,\
                      sFilename_out, \
                      sGrid,\
                      iFlag_scientific_notation_x_in=None,\
                      iFlag_scientific_notation_y_in=None,\
                      iSize_x_in = None, \
                      iSize_y_in = None,  \
                      iDPI_in = None ,\
                      iFlag_log_x_in = None,\
                      iFlag_log_y_in = None,\
                      dMin_x_in = None, \
                      dMax_x_in = None, \
                      dMin_y_in = None, \
                      dMax_y_in = None, \
                      dSpace_x_in = None, \
                      dSpace_y_in = None, \
                      sFormat_x_in =None,\
                      sFormat_y_in =None,\
                      sLabel_x_in =None,\
                      sLabel_y_in = None , \
                      aLabel_legend_in = None,\
                      sTitle_in = None):
    #number of dataset
    nData = len(aData_x)

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_scientific_notation_x_in is not None:
        iFlag_scientific_notation_x = iFlag_scientific_notation_x_in
    else:
        iFlag_scientific_notation_x = 0

    if iFlag_scientific_notation_y_in is not None:
        iFlag_scientific_notation_y = iFlag_scientific_notation_y_in
    else:
        iFlag_scientific_notation_y = 0

    if iFlag_log_x_in is not None:
        iFlag_log_x = iFlag_log_x_in
    else:
        iFlag_log_x = 0

    if iFlag_log_y_in is not None:
        iFlag_log_y = iFlag_log_y_in
    else:
        iFlag_log_y = 0

    if sLabel_x_in is not None:
        sLabel_X = sLabel_x_in
    else:
        sLabel_X = ''

    if sLabel_y_in is not None:
        sLabel_Y = sLabel_y_in
    else:
        sLabel_Y = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )

    left, width = 0.1, 0.75
    bottom, height = 0.1, 0.75
    spacing = 0.005
    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom + height + spacing, width, 0.15]
    rect_histy = [left + width + spacing, bottom, 0.15, height]

    #sns.regplot(x, y, lowess=True)
    #ax_scatter = sns.regplot(x=aData_x, y=aData_y, marker="+", lowess=True)
    ax_scatter = plt.axes(rect_scatter)
    ax_scatter.tick_params(direction='in', top=True, right=True)
    ax_histx = plt.axes(rect_histx)
    ax_histx.tick_params(direction='in', labelbottom=False)
    ax_histy = plt.axes(rect_histy)
    ax_histy.tick_params(direction='in', labelleft=False)


    nPoint = len(aData_x)
    for i in range(nData):
        du= np.array(aData_x[i])
        dv= np.array(aData_y[i])
        du= du.flatten()
        dv =  dv.flatten()
        if (i==0):
            dummyx = du
            dummyy =dv
        else:
            dummyx = np.concatenate([dummyx, du])
            dummyy = np.concatenate([dummyy, dv])
    
    x_min = np.min([dummyx, dummyy])
    x_max = np.max([dummyx, dummyy])
    y_min = np.min([dummyx, dummyy])
    y_max = np.max([dummyx, dummyy])


    if(nData>=3):
        if (nData<=12):
            aColor= create_diverge_rgb_color_hex(nData)
        else:            
            a = rand_hex_color(num=2)
            b = polylinear_gradient(a, nData)
            aColor = b['hex']
            pass
    else:
        if nData==2:
            aColor= ['red','blue']
        else:
            aColor=['red']
    for i in range(nData):

        x = aData_x[i]
        y = aData_y[i]

        #cmap = plt.get_cmap('BuPu')
        ax_scatter.scatter(x, y,  alpha=0.5,color=aColor[i])
        #ax_scatter.set_facecolor('silver')
    
    ax_scatter.axis('on')
    ax_scatter.grid(which='major', color='grey', linestyle='--', axis='y')

    ax_scatter.tick_params(axis="x", labelsize=13)
    ax_scatter.tick_params(axis="y", labelsize=13)

    ax_scatter.set_xmargin(0.05)
    ax_scatter.set_ymargin(0.15)

    ax_scatter.set_xlabel(sLabel_X,fontsize=12)
    ax_scatter.set_ylabel(sLabel_Y,fontsize=12)
    ax_scatter.set_title( sTitle, loc='center', fontsize=15)
    # round to nearest years...

    if sFormat_x_in is not None:
        sFormat_x=sFormat_x_in
        ax_scatter.xaxis.set_major_formatter(ticker.FormatStrFormatter(sFormat_x))

        #ax_scatter.xaxis.set_major_formatter(ticker.FormatStrFormatter('%.1e'))'%.1f'

    if sFormat_y_in is not None:
        sFormat_y = sFormat_y_in
        ax_scatter.yaxis.set_major_formatter(ticker.FormatStrFormatter(sFormat_y))

        #ax_scatter.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1f'))'%.1f'

    ax_scatter.tick_params(axis='y', pad=8)
    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = x_min

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = x_max

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = y_min

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = y_max

    if dSpace_x_in is not None:
        dSpace_x = dSpace_x_in
    else:
        dSpace_x = (dMax_x - dMin_x) / 4.0

    if dSpace_y_in is not None:
        dSpace_y = dSpace_y_in
    else:
        dSpace_y = (dMax_y - dMin_y) /4.0

    ax_scatter.set_xlim( dMin_x, dMax_x )
    ax_scatter.set_ylim( dMin_y, dMax_y)


    ax_scatter.xaxis.set_major_locator(ticker.MaxNLocator(prune='upper', nbins=5))


    if iFlag_log_x ==1:
        aLabel_x = []
        for i in np.arange( dMin_x, dMax_x +1, 1 ):
            sTicklabel = r'$10^{{{}}}$'.format(int(i))
            aLabel_x.append(sTicklabel)
            pass

        ax_scatter.set_xticks(np.arange( dMin_x, dMax_x +1, 1 ))
        ax_scatter.set_xticklabels(aLabel_x)
    else:
        if iFlag_scientific_notation_x ==1:
            formatter = ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            formatter.set_powerlimits((-1,1)) # you might need to change here
            ax_scatter.xaxis.set_major_formatter(formatter)
        else:
            pass


    ax_scatter.yaxis.set_major_locator(ticker.MultipleLocator(base = dSpace_y))


    if iFlag_log_y ==1:
        aLabel_y = []
        for i in np.arange( dMin_y, dMax_y +1, 1 ):
            sTicklabel = r'$10^{{{}}}$'.format(int(i))
            aLabel_y.append(sTicklabel)
            pass

        ax_scatter.set_yticks(np.arange( dMin_y, dMax_y +1, 1 ))
        ax_scatter.set_yticklabels(aLabel_y)
        pass
    else:
        if iFlag_scientific_notation_y ==1:
            formatter = ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            formatter.set_powerlimits((-1,1)) # you might need to change here
            ax_scatter.yaxis.set_major_formatter(formatter)

        pass

    dRatio = (float(iSize_y)/iSize_x) / ( (dMax_y-dMin_y )/ ( dMax_x-dMin_x ) )
    ax_scatter.set_aspect(dRatio)  #this one set the y / x ratio

    line = mlines.Line2D([0, 1], [0, 1], color='red')
    transform = ax_scatter.transAxes
    line.set_transform(transform)
    ax_scatter.add_line(line)
    handles = []  
    # create the corresponding number of labels (= the text you want to display)
    labels = []
    # manually define a new patch 
     
    for i in range(nData):
        handles.append( mpl_patches.Patch( color= aColor[i],  label=aLabel_legend[i])  )
        labels.append(aLabel_legend[i])

    iFlag_lowess = 0
    if(iFlag_lowess==1):      

        y_sm, y_std, order = scatter_lowess(aData_x, aData_y, f=1./3.)
        ax_scatter.plot(x[order], y_sm[order], color='tomato')
        ax_scatter.fill_between(x[order], \
                                y_sm[order] - 1.96*y_std[order], \
                                y_sm[order] + 1.96*y_std[order], \
                                alpha=0.3)

        sLabel_legend_lowess2 = 'LOWESS uncertainty'
      
        labels.append(sLabel_legend_lowess2)

    ax_scatter.legend(handles, labels,\
                      loc="upper right", fontsize=12,\
                      fancybox=True, \
                      framealpha=0.7,\
                      #handlelength=0, \
                      handletextpad=0)

    ax_scatter.tick_params(which='both', # Options for both major and minor ticks
                           top='off', # turn off top ticks
                           left='off', # turn off left ticks
                           right='off',  # turn off right ticks
                           bottom='off') # turn off bottom ticks

    for i in range(nData):
        x = aData_x[i].flatten()
        try:
            if np.max(x) > np.min(x):
                density = gaussian_kde(x)
            else:
                return
        except ValueError:
            print(sGrid,x )
  
        
        xx = np.linspace(dMin_x, dMax_x,1000)
        yy = density(xx)
        ax_histx.plot(xx,yy, color='navy')
        ax_histx.fill_between(xx, yy, 0, linewidth=3,  color = 'lightblue')

        ax_histx.set_xlim( dMin_x, dMax_x )
        ax_histx.set_ylim( 0, auto=None )

        ax_histx.axis('on')
        ax_histx.grid(which='major', color='white', linestyle='-', axis='x')
        ax_histx.xaxis.set_major_locator(ticker.MultipleLocator(base = dSpace_x/2.0))
        ax_histx.spines['right'].set_visible(False)
        ax_histx.spines['top'].set_visible(False)
        ax_histx.spines['bottom'].set_visible(False)
        ax_histx.spines['left'].set_visible(False)
        ax_histx.tick_params(which='both', # Options for both major and minor ticks
                             top='off', # turn off top ticks
                             left='off', # turn off left ticks
                             right='off',  # turn off right ticks
                             bottom='off') # turn off bottom ticks

        ax_histx.axes.get_yaxis().set_visible(False)
        ax_histx.tick_params(axis='x', colors='white')

        #y margin
        y = aData_y[i].flatten()
    
        density = gaussian_kde(y)
        xx = np.linspace(dMin_y, dMax_y,1000)
        yy = density(xx)
        xx, yy = yy, xx
        ax_histy.plot(xx,yy, color='navy')
        ax_histy.fill_betweenx(yy, 0, xx, linewidth=3,  color = 'lightblue')

        ax_histy.set_xlim(0, auto=None)
        ax_histy.set_ylim(dMin_y, dMax_y)

        ax_histy.axis('on')
        ax_histy.grid(which='major', color='white', linestyle='-', axis='y')
        ax_histy.yaxis.set_major_locator(ticker.MultipleLocator(base = dSpace_y/2.0))
        ax_histy.spines['right'].set_visible(False)
        ax_histy.spines['top'].set_visible(False)
        ax_histy.spines['bottom'].set_visible(False)
        ax_histy.spines['left'].set_visible(False)
        ax_histy.axes.get_xaxis().set_visible(False)


        ax_histy.tick_params(axis='y', colors='white')

        ax_histy.tick_params(which='both', # Options for both major and minor ticks
                             top='off', # turn off top ticks
                             left='off', # turn off left ticks
                             right='off',  # turn off right ticks
                             bottom='off') # turn off bottom ticks



    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    print('finished plotting')



import os
import numpy as np
def text_reader_string( sFilename_in,     ncolumn_in = None,      nrow_in = None,      cDelimiter_in = None,      iFlag_remove_quota = None,      iSkipline_in =  None ):
    """
    Read a text based file

    Args:
        sFilename_in (string): The text filename
        ncolumn_in (int, optional): The number of column. Defaults to None.
        nrow_in (int, optional): The number of row. Defaults to None.
        cDelimiter_in (char, optional): The char to seperate a line. Defaults to None.
        iFlag_remove_quota (int, optional): The flag to remove the quota in line string. Defaults to None.
        iSkipline_in (int, optional): The line count to skip. Defaults to None.

    Returns:
        numpy.arry: The data inside the text file.
    """
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    aData_out = -1
    
        
    #check the parameter
    if ncolumn_in is not None:
        iFlag_column = 1
        ncolumn_out = ncolumn_in
    else:
        iFlag_column = 0
        
    if nrow_in is not None :
        #there is nothing
        nrow_out = nrow_in
        #print(' ')
    else :
        #get the total line number
        ifs = open(sFilename_in, "r")
        nrow_out = len(ifs.readlines())
        ifs.close()
    sLine=' '
    ifs = open(sFilename_in, "r")
    if iSkipline_in is not None:
        nrow_out =  nrow_out - iSkipline_in
        for i in range(iSkipline_in):
            ifs.readline()
    else:
        pass
    if iFlag_remove_quota is not None:
        iFlag_iFlag_remove_quota = 1
    else:
        iFlag_iFlag_remove_quota = 0
    #get delimiter
    if cDelimiter_in is not None:
        iFlag_delimiter = 1
    else:
        iFlag_delimiter = 0
        #cDelimiter_in = ' '
        pass
    
    
    if iFlag_delimiter == 1:
        if iFlag_column == 1:
            if ncolumn_out < 1 :
                print( 'The file has no data!' )
                pass
            aData_out = np.full( (nrow_out, ncolumn_out), '  ' , dtype=object )    
            for iRow in range(nrow_out):
                sLine=(ifs.readline()).rstrip()
                if iFlag_iFlag_remove_quota ==1:
                    sLine.replace('"','')
                else:
                    pass    
                aDummy = sLine.split(cDelimiter_in)
                iDummy = len(aDummy)
                if iDummy == ncolumn_out:
                    aData_out[iRow] = aDummy
                else:
                    #something is missing
                    aDummy2  =np.full(ncolumn_out, '  ' , dtype=object)
                    aDummy2[0: iDummy]= aDummy
                    aData_out[iRow]= aDummy2
                    pass
        else :
            sLine=(ifs.readline()).rstrip()
            dummy = sLine.split(cDelimiter_in)
            ncolumn_out = len(dummy)
            #check ncolumn_in count
            if ncolumn_out < 1 :
                print( 'The file has no data!' )  
                pass
            aData_out = np.full( (nrow_out, ncolumn_out), '  ' , dtype=object )         
            aData_out[0] = dummy
            for iRow in range(1, nrow_out):
                sLine=(ifs.readline()).rstrip()
                if iFlag_iFlag_remove_quota ==1:
                    sLine.replace('"','')
                else:
                    pass    
                aDummy = sLine.split(cDelimiter_in)
                iDummy = len(aDummy)
                if iDummy == ncolumn_out:
                    aData_out[iRow] = aDummy
                else:
                    #something is missing
                    aDummy2  =np.full(ncolumn_out, '  ' , dtype=object)
                    aDummy2[0: iDummy]= aDummy
                    aData_out[iRow]= aDummy2
                    pass
            
    else :
        if iFlag_column == 1:
            #check ncolumn_in count
            if ncolumn_out < 1 :                
                return aData_out                      
            aData_out = np.full( (nrow_out, ncolumn_out), '  ', dtype=object )
            for iRow in range(nrow_out):
                dummy1 = ifs.readline()
                #print('New line is: ' + dummy1)
                dummy2 = dummy1.rstrip()
                if iFlag_iFlag_remove_quota ==1:
                    dummy2=dummy2.replace('"','')
                else:
                    pass
                dummy3 = dummy2.split()
                #print(dummy3)
                aData_out[iRow] = dummy3
        else :
            sLine=(ifs.readline()).rstrip()
            dummy = sLine.split()
            ncolumn_out = len(dummy)
            if ncolumn_out < 1 :                
                return aData_out          
            aData_out = np.full( (nrow_out, ncolumn_out), '  ', dtype=object )
            if iFlag_iFlag_remove_quota ==1:
                sLine=sLine.replace('"','')
            else:
                pass
            aData_out[0] =  dummy
            for iRow in range(1,nrow_out):
                dummy1 = ifs.readline()
                #print('New line is: ' + dummy1)
                dummy2 = dummy1.rstrip()
                if iFlag_iFlag_remove_quota ==1:
                    dummy2=dummy2.replace('"','')
                else:
                    pass
                dummy3 = dummy2.split()
                #print(dummy3)
                aData_out[iRow] = dummy3
    #print(aData_out)
    ifs.close()

    

    return aData_out



import os
import sys
import numpy as np
from osgeo import gdal, gdalconst, ogr, osr

def gdal_read_shapefile(sFilename_in):
    """
    Read a ESRI Shapefile. 
    Be careful that GDAL Python blinding has issue with some objects

    Args:
        sFilename_in (string): The filename

    Returns:
        tuple: aFeature, pSpatial_reference, lFeatureCount
    """

    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    sDriverName = "ESRI Shapefile"
    pDriver = ogr.GetDriverByName( sDriverName )

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName)

    pDataset = pDriver.Open(sFilename_in, gdal.GA_ReadOnly)
    
    # Check to see if shapefile is found.
    if pDataset is None:
        print ('Could not open %s' % (sFilename_in))
    else:
        

        pLayer = pDataset.GetLayer()
        lFeatureCount = pLayer.GetFeatureCount()
        print ("Number of features in %s: %d" %  (os.path.basename(sFilename_in),lFeatureCount))
        
        pSpatial_reference = pLayer.GetSpatialRef() 
        
        aFeature=[]
        for pFeature in pLayer:         

            pGeometry = pFeature.GetGeometryRef()
            #print (pGeometry.Centroid().ExportToWkt())
            aFeature.append(pGeometry)       

        return aFeature, pSpatial_reference, lFeatureCount


import os, sys
import numpy as np
from scipy.stats import gaussian_kde
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib.ticker import MaxNLocator
import matplotlib.patches as mpl_patches

def histogram_plot(aData, \
    sFilename_out, \
    iSize_x_in = None, \
    iSize_y_in = None, \
    iDPI_in = None, \
    dMin_in = None, \
    dMax_in = None, \
    dMin_x_in = None, \
    dMax_x_in = None, \
    dSpace_x_in = None, \
    dSpace_y_in = None, \
    sLabel_x_in = None, \
    sLabel_y_in = None, \
    sTitle_in = None):

    """
    Draw a histogram for single dataset
    """

    if iSize_x_in is not None:        
        iSize_x = iSize_x_in
    else:       
        iSize_x = 12
    if iSize_y_in is not None:        
        iSize_y = iSize_y_in
    else:       
        iSize_y = 9
    if iDPI_in is not None:        
        iDPI = iDPI_in
    else:       
        iDPI = 300
    if dMin_in is not None:        
        dMin = dMin_in
    else:       
        dMin = np.min(aData)

    if dMax_in is not None:        
        dMax = dMax_in
    else:       
        dMax = np.max(aData)

    if dMin_x_in is not None:        
        dMin_x = dMin_x_in
    else:       
        dMin_x = dMin

    if dMax_x_in is not None:        
        dMax_x = dMax_x_in
    else:       
        dMax_x = dMax

    if dSpace_x_in is not None:        
        dSpace_x = dSpace_x_in
    else:       
        #it may be calculated 
        pass

    if dSpace_y_in is not None:        
        dSpace_y = dSpace_y_in
    else:       
        #it may be calculated 
        pass

    if sLabel_x_in is not None:        
        sLabel_x = sLabel_x_in
    else:        
        sLabel_x = ''

    if sLabel_y_in is not None:        
        sLabel_y = sLabel_y_in
    else:        
        sLabel_y = ''
    
    if sTitle_in is not None:        
        sTitle = sTitle_in
    else:        
        sTitle = ''
    
    good_index = np.where( (aData >= dMin) & (aData<= dMax)  )
    
    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x )   
    fig.set_figheight( iSize_y )

    left, width = 0.15, 0.7
    bottom, height = 0.1, 0.85
    spacing = 0.005
    rect_histogram = [left, bottom, width, height]
   
    ax_histo = plt.axes(rect_histogram)
    ax_histo.tick_params(direction='in', top=True, right=True)

    aData = aData[good_index]

    ax_histo.hist(aData, int((dMax_x-dMin_x)/dSpace_x))  

    ax_histo.set_xlabel(sLabel_x,fontsize=13 )
    ax_histo.set_ylabel(sLabel_y,fontsize=13 )   
   
  
    ax_histo.set_xlim( dMin_x, dMax_x )
   
    ax_histo.axis('on')   
    ax_histo.grid(which='major', color='white', linestyle='-', axis='y')
    
   
    leg = ax_histo.legend(bbox_to_anchor=(1.0,1.0),loc='upper right', frameon=True)
    
    frame = leg.get_frame()
    frame.set_edgecolor('black')

    ax_histo.set_title(sTitle)
    
    plt.savefig(sFilename_out, bbox_inches='tight')
                       
    plt.close('all')
    print('finished plotting')


def split_string_into_chunk(sString_in,iChunk_size_in=None): 
    """
    Split a string into pieces

    Args:
        sString_in (string): The input strin
        iChunk_size_in (int, optional): the number of chuck. Defaults to None.

    Yields:
        list: Subset of the string
    """
    iTotel_size = len(sString_in)
    if iChunk_size_in is not None:
        iChunk_size = iChunk_size_in
    else:
        iChunk_size = 10
        
    iChunk_count = iTotel_size//iChunk_size
    for pos in range(0, iTotel_size, iChunk_count):
        yield sString_in[pos:pos+iChunk_count]
    


import os
import numpy as np
import xml.etree.ElementTree as ET

def parse_xml_file(sFilename_xml_in):
    """
    Parse an XML file

    Args:
        sFilename_xml_in (string): The XML filename

    Returns:
        dict: The content in the XML file
    """

    if os.path.exists(sFilename_xml_in):
        pass
    else:
        print('The xml file does not exist!')
        return
    tree = ET.parse(sFilename_xml_in)
    root = tree.getroot()
    namelist = {}
    keys_str = []

    for entry in root.findall('./group/entry'):
        key = entry.get('id')
        dtype = entry.find('type').text
        if dtype=='char':
            keys_str.append(key)
        if 'value' in entry.keys():
            # single value setting
            if dtype=='integer':
                value = int(entry.get('value'))
            elif dtype=='real':
                value = float(entry.get('value'))
            elif dtype=='logical':
                if entry.get('value')=='TRUE':
                    value = True
                else:
                    value = False
            else:
                value = entry.get('value')
            namelist[key] = value
        else:
            # value is an array
            values = []
            for value in entry.findall('./values'):
                key = entry.get('id')
                dtype = entry.find('type').text
                if dtype=='integer':
                    values.append(int(value.text))
                elif dtype=='real':
                    values.append(float(value.text))
                elif dtype=='logical':
                    if value.text=='TRUE':
                        values.append(True)
                    else:
                        values.append(False)
                else:
                    values.append(value.text)
            namelist[key] = np.array(values, dtype=np.float64, order='F')

    # fill environment variable values
    for key in keys_str:
        keys_str_iter = list(set(keys_str)-set([key]))
        for okey in keys_str_iter:
            string = namelist[okey]
            if string.find('$'+key)>=0:
                new_string = string.replace('$'+key, namelist[key])
                namelist[okey] = new_string
                
    return namelist


import os, sys
import numpy as np
import osgeo
from osgeo import ogr, osr, gdal, gdalconst

gdal.UseExceptions()    # Enable exceptions


def reproject_coordinates(dx_in, dy_in, pSpatial_reference_source_in, pSpatial_reference_target_in=None):
    """ Reproject a pair of x,y coordinates. 

    Args:
        dx_in (float): X Coordinate of point
        dy_in (float): Y Coordinate of
        pSpatial_reference_source_in (osr): The source spatial reference of point
        pSpatial_reference_target_in (osr, optional): The target spatial reference of point. Defaults to None.

    Returns:
        Tuple: dx_new, dy_new
    """

    if pSpatial_reference_target_in is not None:

        pass
    else:
        pSpatial_reference_target_in = osr.SpatialReference()
        pSpatial_reference_target_in.ImportFromEPSG(4326)
        
        pass

    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        pSpatial_reference_source_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        pSpatial_reference_target_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( pSpatial_reference_source_in, pSpatial_reference_target_in)
   
    dx_out,dy_out, z = pTransform.TransformPoint( dx_in,dy_in)
    
    return dx_out, dy_out

def reproject_coordinates_batch(aX_in, aY_in, pSpatial_reference_source_in, pSpatial_reference_target_in=None):
    """ Reproject a list of x, y coordinates.

    Args:
        aX_in (list): A list of X Coordinate of points
        aY_in (list): A list of Y Coordinate of points
        pSpatial_reference_source_in (osr): The source spatial reference of point
        pSpatial_reference_target_in (osr, optional): The target spatial reference of point. Defaults to None.

    Returns:
        Tuple: aX_out, aY_out
    """

    if pSpatial_reference_target_in is not None:

        pass
    else:
        pSpatial_reference_target_in = osr.SpatialReference()
        pSpatial_reference_target_in.ImportFromEPSG(4326)
        
        pass

    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        pSpatial_reference_source_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        pSpatial_reference_target_in.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( pSpatial_reference_source_in, pSpatial_reference_target_in)

    npoint = len(aX_in)
    aX_out=list()
    aY_out=list()
    for i in range(npoint):
        x0 = aX_in[i]
        y0 = aY_in[i]
   
        x1,y1, z = pTransform.TransformPoint( x0,y0)

        aX_out.append(x1)
        aY_out.append(y1)
    
    return aX_out,aY_out

def obtain_raster_metadata_geotiff(sFilename_geotiff_in):
    """retrieve the metadata of a geotiff file

    Args:
        sFilename_geotiff (string): The filename of geotiff

    Returns:
       Tuple: dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, pSpatial_reference, pProjection, pGeotransform
    """

    if os.path.exists(sFilename_geotiff_in):
        pass
    else:
        print('The file does not exist!')
        return
    
    sDriverName='GTiff'    
    pDriver = gdal.GetDriverByName(sDriverName)
    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 
   
    pDataset = gdal.Open(sFilename_geotiff_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_geotiff_in)
        sys.exit("Try again!")
    else: 
        pProjection = pDataset.GetProjection()
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)
    
    
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        #nband = pDataset.RasterCount

        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        dPixelWidth = pGeotransform[1]
        pPixelHeight = pGeotransform[5]       
        
        
        return dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, pSpatial_reference, pProjection, pGeotransform

def obtain_shapefile_metadata(sFilename_shapefile_in):
    """
    Obtain the metadata of a shapefile

    Args:
        sFilename_shapefile (string): The filename of the shapefile

    Returns:
        Tuple: left_min, right_max, bot_min, top_max
    """
    if os.path.exists(sFilename_shapefile_in):
        pass
    else:
        print('The  shapefile does not exist!')
        return


    sDriverName='ESRI Shapefile'    
    pDriver_shapefile = ogr.GetDriverByName(sDriverName)
    if pDriver_shapefile is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 
   
    pDataset = pDriver_shapefile.Open(sFilename_shapefile_in, gdal.GA_ReadOnly)
    pLayer = pDataset.GetLayer(0)
  

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_shapefile_in)
        sys.exit("Try again!")
    else:    
        
        iFlag_first=1
    
        for feature in pLayer:
            pGeometry = feature.GetGeometryRef()
            pEnvelope = pGeometry.GetEnvelope()

            if iFlag_first ==1:
                left_min = pEnvelope[0]
                right_max =  pEnvelope[1]
                bot_min =  pEnvelope[2]
                top_max =  pEnvelope[3]
                iFlag_first = 0
            else:
                left_min = np.min([left_min,  pEnvelope[0]])
                right_max = np.max([right_max,  pEnvelope[1]])
                bot_min = np.min([bot_min,  pEnvelope[2]])
                top_max = np.max([top_max,  pEnvelope[3]])

        return left_min, right_max, bot_min, top_max




def create_diverge_rgb_color_hex(ncolor, iFlag_reverse_in = None):
    """
    choose diverge color from: https://colorbrewer2.org/
    """

    if ncolor < 3 or ncolor > 12:
        return -1
    else:
        if ncolor == 3:
            colors_hex = [  '#fc8d59',\
                                '#ffffbf',\
                                '#91bfdb']

        else:
            if ncolor == 4:
                colors_hex = [  '#d7191c',\
                                '#fdae61',\
                                '#abdda4',\
                                '#2b83ba']
            if ncolor == 5:
                colors_hex = [  '#d7191c',\
                                '#fdae61',\
                                '#ffffbf',\
                                '#abdda4',\
                                '#2b83ba']
            if ncolor == 6:
                colors_hex = [  '#66c2a5',\
                                '#fc8d62',\
                                '#8da0cb',\
                                '#e78ac3',\
                                '#a6d854',\
                                '#ffd92f' ]

                #another option
                colors_hex = [  '#e41a1c',\
                                '#377eb8',\
                                '#4daf4a',\
                                '#984ea3',\
                                '#ff7f00',\
                                '#ffff33']
                
            if ncolor == 7:
                colors_hex = [  '#d73027', \
                                '#fc8d59', \
                                '#fee090', \
                                '#ffffbf', \
                                '#e0f3f8', \
                                '#91bfdb', \
                                '#4575b4'  ]
            if ncolor == 8:
                colors_hex = [  '#d73027', \
                                '#fc8d59', \
                                '#fee090', \
                                '#ffffbf', \
                                '#e0f3f8', \
                                '#91bfdb', \
                                '#4575b4'  ]
            if ncolor == 9:
                colors_hex = [  '#d53e4f',\
                                '#f46d43',\
                                '#fdae61',\
                                '#fee08b',\
                                '#ffffbf',\
                                '#e6f598',\
                                '#abdda4',\
                                '#66c2a5',\
                                '#3288bd']
            if ncolor == 10:
                colors_hex = [  '#9e0142', \
                                '#d53e4f', \
                                '#f46d43', \
                                '#fdae61', \
                                '#fee08b', \
                                '#e6f598', \
                                '#abdda4', \
                                '#66c2a5', \
                                '#3288bd', \
                                '#5e4fa2']
            else:
                pass
                
        #add the reverse feature
        if iFlag_reverse_in is not None:
            colors_hex.reverse()
        else:
            pass
        return colors_hex


import os
def envi_write_header(sFilename_in, aHeader_in):
    """
    write the header for ENVI raster file from the dictionary data type, 
    currently it only supports WGS84 and a dict is required to store all the information. 
    https://www.harrisgeospatial.com/docs/ENVIGridDefinition.html

    Parameters
    ----------
        sFilename_in (string): The filename
        aHeader (dict): A dict that contains all required information
    Returns
    -------

    See Also
    --------
    
    Examples
    --------
    """
    
    
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The envi file does not exist!')
        return

    pFile = open(sFilename_in, 'w')
    sLine = 'ENVI' + '\n'
    pFile.write(sLine)
    sLine = 'description = ' + aHeader_in['sFilename'] + '\n'
    pFile.write(sLine)
    sLine = 'samples = '  + aHeader_in['ncolumn'] + '\n'
    pFile.write(sLine)
    sLine = 'lines = ' + aHeader_in['nrow'] + '\n'
    pFile.write(sLine)
    sLine = 'bands = ' + aHeader_in['nband'] + '\n'
    pFile.write(sLine)
    sLine = 'header offset = ' + aHeader_in['offset'] + '\n'
    pFile.write(sLine)
    sLine = 'data type = ' + aHeader_in['data_type'] + '\n'
    pFile.write(sLine)
    sLine = 'interleave = ' + aHeader_in['bsq'] + '\n'
    pFile.write(sLine)
    sLine = 'sensor type = Unknown' + '\n'
    pFile.write(sLine)
    sLine = 'byte order = ' + aHeader_in['byte_order'] + '\n'
    pFile.write(sLine)
    sLine = 'data ignore value = ' + aHeader_in['missing_value'] + '\n'
    pFile.write(sLine)

    #only one spatial reference is supported here.
    sLine = 'map info = {Geographic Lat/Lon, 1.000, 1.000, ' \
        + aHeader_in['ULlon'] + ', ' + aHeader_in['ULlat'] + ', ' \
        + aHeader_in['pixelSize'] + ', ' + aHeader_in['pixelSize'] + ', ' \
        + 'WGS-84, units = Degrees}' + '\n'
    pFile.write(sLine)

    sLine = 'wavelength units = Unknown ' + '\n'
    pFile.write(sLine)
    pFile.close()



import os, sys
from datetime import datetime
import numpy as np
import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.ticker as ticker

import pandas as pd
from statsmodels.tsa.seasonal import STL
from statsmodels.tsa.stattools import adfuller

from pyearth.system.define_global_variables import *

from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex

def plot_time_series_analysis(aTime, \
                              aData, \
                              sFilename_out,\
                              sVariable,\
                              iDPI_in = None,\
                              iFlag_without_raw_in= None,\
                              iFlag_log_in = None,\
                              iReverse_y_in = None, \
                              iSize_x_in = None, \
                              iSize_y_in = None, \
                              dMax_x_in = None, \
                              dMin_x_in = None, \
                              dMax_y_in = None, \
                              dMin_y_in = None, \
                              dSpace_x_in = None,\
                              dSpace_y_in = None,\
                              aMarker_in = None,\
                              aColor_in = None,\
                              aLinestyle_in = None,\
                              sLabel_x_in = None, \
                              sLabel_y_in = None, \
                              aLabel_legend_in = None,\
                              sDate_type_in = None,\
                              sFormat_y_in =None,\
                              sTitle_in = None):
    #find how many data will be plotted


    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_without_raw_in is not None:
        iFlag_without_raw = iFlag_without_raw_in
    else:
        iFlag_without_raw = 0

    if iReverse_y_in is not None:
        iReverse_y = iReverse_y_in
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = sVariable

    nData = 4
    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.array([ sVariable, 'Trend','Season','Residual' ])

    if sTitle_in is not None:
        sTitle = sTitle_in.title()
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker = ['o','.','*','+']

    if aColor_in is not None:
        aColor = aColor_in
    else:
        if(nData>=3):
            aColor= create_diverge_rgb_color_hex(nData)
        else:
            if nData==2:
                aColor= ['red','blue']
            else:
                aColor=['red']

    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle =  ['-','--','-.' ,'solid']

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.datetime64(np.nanmax(aTime), 'Y')

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.datetime64(np.nanmin(aTime), 'Y')

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = np.nanmax(aData) * 1.2

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = np.nanmin(aData) * 0.8 #if it has negative value, change here
    if (dMax_y <= dMin_y ):
        return

    if dSpace_y_in is not None:
        iFlag_space_y =1
        dSpace_y = dSpace_y_in
    else:
        iFlag_space_y=0
        pass

    adf_test = adfuller(aData)
    #print(adf_test)
    print("ADF = " + str(adf_test[0]))
    print("p-value = " +str(adf_test[1])  )

    #fig = plt.figure( dpi=iDPI )
    #fig.set_figwidth( iSize_x )
    #fig.set_figheight( iSize_y )

    #pAxGrid = AxesGrid(fig, 111,
    #                    nrows_ncols=(4,1),
    #                    axes_pad=0.6,
    #                    label_mode='')  # note the empty label_mode


    pYear = mdates.YearLocator(1)   # every year
    pMonth = mdates.MonthLocator()  # every month
    if sDate_type_in is not None:
        if sDate_type_in == 'month':
            pMonth = mdates.MonthLocator(3)
        else:
            print(sDate_type_in)
    else:
        print(sDate_type_in)
        pass

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0

    sYear_format = mdates.DateFormatter('%Y')

    aData_tsa = pd.Series(aData, index=pd.date_range(aTime[0], \
                                                     periods=len(aTime), freq='M'), name = sVariable)

    
    #https://www.statsmodels.org/stable/generated/statsmodels.tsa.seasonal.STL.html#statsmodels.tsa.seasonal.STL
    stl = STL(aData_tsa, seasonal=13)
    aTSA = stl.fit()
    #part 1
    #plot time series



    if iFlag_without_raw == 1:
        aData_all = [aTSA.trend, aTSA.seasonal, aTSA.resid ]
        iSize_y = iSize_y * 0.75
        fig, pAxGrid = plt.subplots(nrows= len(aData_all), \
                                    figsize=(iSize_x, iSize_y),\
                                    sharex=True,  dpi=iDPI)
    else:
        aData_all = [aData, aTSA.trend, aTSA.seasonal, aTSA.resid ]
        fig, pAxGrid = plt.subplots(nrows= len(aData_all), \
                                    figsize=(iSize_x, iSize_y),\
                                    sharex=True,  dpi=iDPI)

    for i, ax in enumerate(pAxGrid):


        #ax.set_facecolor('#eafff5')
        if iFlag_without_raw == 1:
            ax.plot( aTime, aData_all[i], \
                     color = aColor[i+1], linestyle = aLinestyle[i+1] ,\
                     marker = aMarker[i+1] ,\
                     label = aLabel_legend[i+1],\
                     zorder=3)

        else:
            ax.plot( aTime, aData_all[i], \
                     color = aColor[i], linestyle = aLinestyle[i] ,\
                     marker = aMarker[i] ,\
                     label = aLabel_legend[i],\
                     zorder=3)

        if  iFlag_without_raw == 0:
            if i==0:
                if iReverse_y ==1:
                    ax.set_ylim( dMin_y,dMax_y  )

                ax.set_title(sTitle,fontsize=13)

                if iFlag_log_in ==1:
                    #we need to change the ticklabel
                    aLabel_y = []
                    for j in np.arange( dMin_y, dMax_y + 1, 1 ):
                        sTicklabel = r'$10^{{{}}}$'.format(int(j))
                        aLabel_y.append(sTicklabel)
                        pass

                    ax.set_yticks(np.arange( dMin_y, dMax_y +1, 1 ))
                    ax.set_yticklabels(aLabel_y)
                    pass
                pass
        else:
            if i==0:
                ax.set_title(sTitle,fontsize=13)


        if iFlag_without_raw == 1:
            ax.set_ylabel(aLabel_legend[i+1],fontsize=12)
            if i == 2:
                ax.plot((dMin_x, dMax_x), (0, 0), color='#000000', linestyle=':', zorder=2)

                ax.set_xlabel('Year',fontsize=12)

        else:
            ax.set_ylabel(aLabel_legend[i],fontsize=12)
            if i == 3:
                ax.plot((dMin_x, dMax_x), (0, 0), color='#000000', linestyle=':', zorder=2)
                ax.set_xlabel('Year',fontsize=12)


        ax.grid(which='major', color='lightgrey', linestyle=':', axis='y', zorder=1)


        ax.set_xlim(dMin_x, dMax_x)
        #ax.xaxis.set_major_locator(pYear)
        #ax.xaxis.set_minor_locator(pMonth)


    #we can now rewrite the plot function here





    #save the result
    #plt.show()
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    #print('finished plotting')



import os, sys
from datetime import datetime
import numpy as np
import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.ticker as ticker


from pyearth.system.define_global_variables import *

from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.visual.color.choose_n_color import polylinear_gradient, rand_hex_color

def plot_time_series_data(aTime_all, \
                        aData_all, \
                          sFilename_out,\
                          iDPI_in = None,\
                          iFlag_log_in = None,\
                          iFlag_scientific_notation_in = None,\
                          ncolumn_in = None,\
                          aFlag_trend_in = None, \
                          iReverse_y_in = None, \
                          iSize_x_in = None, \
                          iSize_y_in = None, \
                          dMax_x_in = None, \
                          dMin_x_in = None, \
                          dMax_y_in = None, \
                          dMin_y_in = None, \
                          dSpace_y_in = None,\
                          aMarker_in = None,\
                          aColor_in = None,\
                          aLinestyle_in = None,\
                          sLabel_y_in = None, \
                          aLabel_legend_in = None,\
                          aLocation_legend_in =None,\
                          sDate_type_in = None,\
                          sFormat_y_in =None,\
                          sLocation_legend_in=None,\
                          sTitle_in = None):
    #find how many data will be plotted
    pShape = aData_all.shape
  
    nData = pShape[0]

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_log_in is not None:
        iFlag_log = 1
    else:
        iFlag_log = 0

    if iFlag_scientific_notation_in is not None:
        iFlag_scientific_notation = 1
    else:
        iFlag_scientific_notation = 0


    if aFlag_trend_in is not None:
        aFlag_trend = aFlag_trend_in
    else:
        aFlag_trend = np.full(nData, 0)

    if iReverse_y_in is not None:
        iReverse_y = iReverse_y_in
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.full(nData,'')

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker=np.full(nData, '+')

    if aColor_in is not None:
        aColor = aColor_in
    else:
        if(nData>=3):
            if (nData<=12):
                aColor= create_diverge_rgb_color_hex(nData)
            else:            
                a = rand_hex_color(num=2)
                b = polylinear_gradient(a, nData)
                aColor = b['hex']
                pass
        else:
            if nData==2:
                aColor= ['red','blue']
            else:
                aColor=['red']

    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle = np.full(nData, '-')

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.datetime64(np.nanmax(aTime_all), 'Y')
    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.datetime64(np.nanmin(aTime_all), 'Y')

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = np.nanmax(aData_all) #* 1.2

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = np.nanmin(aData_all) #if it has negative value, change here

    if (dMax_y <= dMin_y ):
        return

    if dSpace_y_in is not None:
        iFlag_space_y =1
        dSpace_y = dSpace_y_in
    else:
        iFlag_space_y =1
        dSpace_y = (dMax_y - dMin_y) /4.0
        pass

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4] )

    nYear = int( (dMax_x-dMin_x)/ 5 )
    pYear = mdates.YearLocator(nYear)   # every year
    pMonth = mdates.MonthLocator()  # every month

    if sDate_type_in is not None:
        if sDate_type_in == 'month':
            pMonth = mdates.MonthLocator(3)
        else:
            print(sDate_type_in)
    else:
        print(sDate_type_in)
        pass

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend=(1.0,1.0)

    if ncolumn_in is not None:
        ncolumn = ncolumn_in
    else:
        ncolumn = 1

    sYear_format = mdates.DateFormatter('%Y')

    #start loop for each data
    for i in np.arange(1, nData+1):

        x1 = aTime_all[i-1]
        y1 = aData_all[i-1]
        ax.plot( x1, y1, \
                 color = aColor[i-1], linestyle = aLinestyle[i-1] ,\
                 marker = aMarker[i-1] ,\
                 label = aLabel_legend[i-1])

        #calculate linear regression
        iFlag_trend = aFlag_trend[i-1]
        if iFlag_trend ==1:
            nan_index = np.where(y1 == missing_value)
            y1[nan_index] = np.nan
            good_index = np.where(  ~np.isnan(y1))

            x_dummy = np.array( [i.timestamp() for i in x1 ] )
            x_dummy = x_dummy[good_index]
            y_dummy = y1[good_index]
            coef = np.polyfit(x_dummy,y_dummy,1)
            poly1d_fn = np.poly1d(coef)
            mn=np.min(x_dummy)
            mx=np.max(x_dummy)
            x2=[mn,mx]
            y2=poly1d_fn(x2)
            x2 = [datetime.fromtimestamp(i) for i in x2 ]
            ax.plot(x2,y2, color = 'orange', linestyle = '-.',  linewidth=0.5)

    ax.axis('on')
    ax.grid(which='major', color='grey', linestyle='--', axis='y')

    #ax.set_aspect(dRatio)  #this one set the y / x ratio

    ax.xaxis.set_major_locator(pYear)
    ax.xaxis.set_minor_locator(pMonth)
    ax.xaxis.set_major_formatter(sYear_format)
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)

    ax.set_xmargin(0.05)
    ax.set_ymargin(0.15)

    ax.set_xlabel('Year',fontsize=12)

    ax.set_title( sTitle, loc='center', fontsize=15)

    ax.set_xlim(dMin_x, dMax_x)

    #next Y axis
    ax.set_ylabel(sLabel_y,fontsize=12)

    if (iReverse_y ==1): #be careful here
        ax.set_ylim( dMax_y, dMin_y )
    else:
        ax.set_ylim( dMin_y, dMax_y )

    if iFlag_log ==1:
        #we need to change the ticklabel
        aLabel_y = []
        nlabel = int( (dMax_y- dMin_y) / dSpace_y) +1
        for i in np.arange( 0, nlabel, 1 ):
            ii = dMin_y + i * dSpace_y
            sTicklabel = r'$10^{{{}}}$'.format(ii)
            aLabel_y.append(sTicklabel)
            pass
        ticks = np.arange( 0, nlabel, 1 ) * dSpace_y + dMin_y
        ax.set_yticks( ticks)
        ax.set_yticklabels(aLabel_y)
        #if (iFlag_format_y ==1):
        #    ax.yaxis.set_major_formatter(ticker.FormatStrFormatter( sFormat_y ) )

        
           
        #ax.yaxis.set_major_locator(ticker.MultipleLocator(dSpace_y))
        ax.yaxis.set_minor_locator(ticker.AutoMinorLocator())

        
        
    else:
        #not log

        if iFlag_scientific_notation ==1:
            formatter = ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            formatter.set_powerlimits((-1,1)) # you might need to change here
            ax.yaxis.set_major_formatter(formatter)
            #most time, when you use scientific notation, you may not need set the space,
            #but you may still set it using the method below

            pass
        else:
            if (iFlag_format_y ==1):
                ax.yaxis.set_major_formatter(ticker.FormatStrFormatter( sFormat_y ) )

            if (iFlag_space_y ==0):
                #ax.yaxis.set_major_locator(ticker.AutoLocator())
                #ax.yaxis.set_minor_locator(ticker.AutoMinorLocator())
                ax.yaxis.set_major_locator(ticker.MaxNLocator(prune='upper', nbins=5))
            else:
                ax.yaxis.set_major_locator(ticker.MultipleLocator(dSpace_y))
                ax.yaxis.set_minor_locator(ticker.AutoMinorLocator())

            pass

    if len(aLabel_legend[0]) > 0:
        ax.legend(bbox_to_anchor=aLocation_legend, \
              loc=sLocation_legend, \
              fontsize=12, \
              ncol= ncolumn)

    #save the result
    #plt.show()
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    #print('finished plotting')



import os, sys

import numpy as np
import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.ticker as ticker

from pyearth.system.define_global_variables import *

from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex



def barplot_data(aData_in,\
                 aLabel_x_in,\
                 aLabel_y_in,\
                 sFilename_out,\
                 iDPI_in = None,\
                 ncolumn_in = None,\
                 iSize_x_in = None, \
                 iSize_y_in = None, \
                 dMax_y_in = None, \
                 dMin_y_in = None, \
                 dSpace_y_in = None,\
                 aMarker_in = None,\
                 aColor_in = None,\
                 aHatch_in = None,\
                 sLabel_y_in = None, \
                 aLabel_legend_in = None,\
                 aLocation_legend_in =None,\
                 sFormat_y_in =None,\
                 sLocation_legend_in=None,\
                 sTitle_in = None):
    """
    Draw a barplot
    """

    aData_in = np.array(aData_in)
    pShape = aData_in.shape

    nData = pShape[0] #len(aLabel_x_in) this maybe checked 

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = np.nanmax(aData_in) * 1.0

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = np.nanmin(aData_in) 

    if (dMax_y <= dMin_y ):
        return

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend=(1.0,1.0)

    if ncolumn_in is not None:
        ncolumn = ncolumn_in
    else:
        ncolumn = 1

    if aColor_in is not None:
        aColor = aColor_in
    else:
        if(nData>=3):
            aColor= create_diverge_rgb_color_hex(nData)
        else:
            if nData==2:
                aColor= ['red','blue']
            else:
                aColor=['red']

    if aHatch_in is not None:
        aHatch = aHatch_in
    else:
        aHatch= np.fill(nData, '+')

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4] )
    x = np.arange(  len(aLabel_x_in)  )
    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel(sLabel_y)
    ax.set_title(sTitle)
    ax.set_xticks(x)
    ax.set_xticklabels(aLabel_x_in)

    total_width = 0.7
    width = total_width / nData

    for i in np.arange(1, nData + 1, 1):

        data1 = aData_in[i-1]
        #need bettern control here, hatch (texture) can be used to betten represet some data
        if i == 1 or i ==2:
            rects = ax.bar( x - total_width * 0.5 + i * width, data1, width, label= aLabel_y_in[i-1],\
                            color = aColor[i-1], hatch = aHatch[i-1], edgecolor = "k")
        else:
            rects = ax.bar( x - total_width * 0.5 + i * width, data1, width, label= aLabel_y_in[i-1],\
                            color = aColor[i-1])#, hatch = aHatch[i-1], edgecolor = "k")
      
        pass

    if (iFlag_format_y ==1):
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter( sFormat_y ) )

    ax.set_ylim( dMin_y, dMax_y )
    ax.legend(bbox_to_anchor=aLocation_legend, \
              loc=sLocation_legend, \
              fontsize=12, \
              ncol= ncolumn)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()


    return



import os, sys
from osgeo import gdal

def gdal_write_envi_file(sFilename_in, aData_in, dPixelWidth_in,dOriginX_in,  dOriginY_in, dMissing_value_in,   pSpatial_reference_in ):
    """
    Write a ENVI standard format raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """
    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass
    

    sDriverName='ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 
    

    nrow, ncolumn = aData_in.shape
    nband = 1

    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in, \
                              ncolumn, \
                              nrow, \
                              nband,\
                              gdal.GDT_Float32)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in, \
                              dPixelWidth_in, \
                              0.0, \
                              dOriginY_in, \
                              0.0, \
                              -dPixelWidth_in])

    pProjection= pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    #Write raster datasets
    pBand = pDataset.GetRasterBand(1)
    pBand.WriteArray(aData_in)
    pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver =None
    pDataset=None
    pBand=None

    return pGeotransform_out, pProjection_out



def gdal_write_envi_file_multiple_band(sFilename_in, aData_in,  dPixelWidth_in,  dOriginX_in,  dOriginY_in,  dMissing_value_in,  pSpatial_reference_in ):
    """
    Write a multi-band ENVI raster file

    Args:
        sFilename_in (string): The filename
        aData_in (numpy.array): The data
        dPixelWidth_in (float): The resolution
        dOriginX_in (float): The location of origin x
        dOriginY_in (float): The location of origin y
        dMissing_value_in (float): The missinge value
        pSpatial_reference_in (osr): The spatial reference

    Returns:
        Tuple: pGeotransform_out, pProjection_out
    """


    if os.path.exists(sFilename_in):
        os.remove(sFilename_in)
        pass

    sDriverName='ENVI'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName) 

    nband, nrow, ncolumn = aData_in.shape


    # Creates a new raster data source
    pDataset = pDriver.Create(sFilename_in, \
                              ncolumn, \
                              nrow, \
                              nband,\
                              gdal.GDT_Float32)

    # Write metadata

    pDataset.SetGeoTransform([dOriginX_in, \
                              dPixelWidth_in, \
                              0.0, \
                              dOriginY_in, \
                              0.0, \
                              -dPixelWidth_in])

    pProjection= pSpatial_reference_in.ExportToPrettyWkt()
    pDataset.SetProjection(pProjection)

    #Write raster datasets
    for iBand in range(nband):
        pBand = pDataset.GetRasterBand(iBand + 1)
        pBand.WriteArray(aData_in[iBand, :, :])
        pBand.SetNoDataValue(dMissing_value_in)

    pGeotransform_out = pDataset.GetGeoTransform()
    pProjection_out = pDataset.GetProjection()

    pDataset.FlushCache()  # Write to disk.
    pDriver =None
    pDataset=None
    pBand=None

    return  sFilename_in, pGeotransform_out, pProjection_out



import os, sys
import numpy as np
import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.ticker as ticker
from datetime import datetime



from pyearth.system.define_global_variables import *
from pyearth.visual.plot.calculate_ticks_space import calculate_ticks_space

def plot_time_series_data_monthly_fill(aTime, aData, \
    sFilename_out,\
    iDPI_in = None,\
    iFlag_trend_in = None, \
    iReverse_Y_in = None, \
    iSize_X_in = None, \
    iSize_Y_in = None, \
    dMax_Y_in =None, \
    dMin_Y_in = None, \
    sMarker_in =None,\
    sLabel_Y_in = None, \
    sLabel_legend_in = None,\
    sTitle_in = None):

    if iDPI_in is not None:        
        iDPI = iDPI_in
    else:       
        iDPI = 300
    
    if iFlag_trend_in is not None:
        iFlag_trend = 1
    else:
        iFlag_trend = 0
        
    if iReverse_Y_in is not None:
        iReverse_Y = 1
    else:
        iReverse_Y = 0

    if iSize_X_in is not None:        
        iSize_X = iSize_X_in
    else:       
        iSize_X = 12
    if iSize_Y_in is not None:        
        iSize_Y = iSize_Y_in
    else:       
        iSize_Y = 9
    
    if sLabel_Y_in is not None:        
        sLabel_Y = sLabel_Y_in
    else:        
        sLabel_Y = ''
    if sLabel_legend_in is not None:        
        sLabel_legend = sLabel_legend_in
    else:        
        sLabel_legend = ''
    if sTitle_in is not None:        
        sTitle = sTitle_in
    else:        
        sTitle = ''

    if sMarker_in is not None:        
        sMarker = sMarker_in
    else:        
        sMarker = '+'

    nstress = len(aTime)
    nan_index = np.where(aData == missing_value)
    aData[nan_index] = np.nan
    good_index = np.where(  ~np.isnan(aData))
    
    if dMax_Y_in is not None:        
        dMax_Y = dMax_Y_in
    else:         
        dMax_Y = np.nanmax(aData) * 1.2 
    if dMin_Y_in is not None:        
        dMin_Y = dMin_Y_in
    else:
        dMin_Y = np.nanmin(aData) #if it has negative value, change here  
    if (dMax_Y <= dMin_Y ):
        return

    

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_X)   
    fig.set_figheight( iSize_Y)         
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4] )  
    pYear = mdates.YearLocator(5)   # every year
    pMonth = mdates.MonthLocator()  # every month
    sYear_format = mdates.DateFormatter('%Y')
    x1 = aTime    
    y1 = (aData[1])[0]
    y_top = (aData[0])[0]
    y_bot = (aData[2])[0]
    ax.fill_between(x1, y_top, y_bot,  facecolor='cornflowerblue')
    ax.plot( x1, y1, \
             color = 'red', linestyle = '--' ,\
                  marker=sMarker, markeredgecolor='blue' ,\
                       label= sLabel_legend)
                       
    #calculate linear regression
    if iFlag_trend ==1: 
        x_dummy = np.array( [i.timestamp() for i in x1 ] )        
        x_dummy = x_dummy[good_index]
        y_dummy = y1[good_index]
        coef = np.polyfit(x_dummy,y_dummy,1)
        poly1d_fn = np.poly1d(coef) 
        mn=np.min(x_dummy)
        mx=np.max(x_dummy)
        x2=[mn,mx]
        y2=poly1d_fn(x2)
        x2 = [datetime.fromtimestamp(i) for i in x2 ]  
        ax.plot(x2,y2, color = 'orange', linestyle = '-.',  linewidth=0.5)

    ax.axis('on')          
    ax.grid(which='major', color='grey', linestyle='--', axis='y') 
    #ax.grid(which='minor', color='#CCCCCC', linestyle=':') #only y axis grid is 
    
    #ax.set_aspect(dRatio)  #this one set the y / x ratio
    ax.xaxis.set_major_locator(pYear)        
    ax.xaxis.set_minor_locator(pMonth)
    ax.xaxis.set_major_formatter(sYear_format)
    ax.tick_params(axis="x", labelsize=10) 
    ax.tick_params(axis="y", labelsize=10)
    
    ax.set_xmargin(0.05)
    ax.set_ymargin(0.15)
    
    ax.set_xlabel('Year',fontsize=12)
    ax.set_ylabel(sLabel_Y,fontsize=12)
    ax.set_title( sTitle, loc='center', fontsize=15)
    # round to nearest years...
    x_min = np.datetime64(aTime[0], 'Y')
    x_max = np.datetime64(aTime[nstress-1], 'Y') + np.timedelta64(1, 'Y')
    ax.set_xlim(x_min, x_max)
    if dMax_Y < 1000 and dMax_Y > 0.1:
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1f'))
    else: 
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.1e'))
    dummy = calculate_ticks_space([y1,y_top,y_bot], nstep_in =5, iFlag_small_in= 1)
    dSpace = dummy[0]
    if(dSpace<= 0):
        ax.invert_yaxis()

    else:
        ax.yaxis.set_major_locator(ticker.MultipleLocator(dSpace))
        dMin_Y = dummy[1]
        dMax_Y = dummy[2]
        if (iReverse_Y ==1):
            ax.set_ylim( dMax_Y, dMin_Y )
        else:
            ax.set_ylim( dMin_Y, dMax_Y )
    ax.legend(bbox_to_anchor=(1.0,1.0), loc="upper right",fontsize=12)
    plt.savefig(sFilename_out, bbox_inches='tight')
                       
    plt.close('all')
    plt.clf()
    #print('finished plotting')
